#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <GL/glut.h>
#include "FrameXform.h"
#include "WaveFrontOBJ.h"

#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795
#endif

double cameras[5][9] = 
{
  {28,18,28, 0,2,0, 0,1,0},   
  {28,18,-28, 0,2,0, 0,1,0}, 
  {-28,18,28, 0,2,0, 0,1,0}, 
  {-12,12,0, 0,2,0, 0,1,0},  
  {0,100,0,  0,0,0, 1,0,0}
};
int cameraCount = sizeof( cameras ) / sizeof( cameras[0] );

int cameraIndex, camID;
vector<FrameXform> wld2cam, cam2wld; 
WaveFrontOBJ* cam;
FrameXform cow2wld;
WaveFrontOBJ* cow;
int cowID;
FrameXform bet2wld;
WaveFrontOBJ* bet;
int betID;
unsigned floorTexID;
int frame = 0;
int width, height;
int selectMode, oldX, oldY;

void drawFrame(float len);

//------------------------------------------------------------------------------
void munge( int x, double& r, double& g, double& b)
{
  r = (x & 255)/double(255);
  g = ((x >> 8) & 255)/double(255);
  b = ((x >> 16) & 255)/double(255);
}

//------------------------------------------------------------------------------
int unmunge( double r, double g, double b)
{
  return (int(r) + (int(g) << 8) + (int(b) << 16));
}

//------------------------------------------------------------------------------
void setCamera()
{
  int i;
  if (frame == 0)
  {
    // intialize camera model
    cam = new WaveFrontOBJ("camera.obj");
    camID = glGenLists(1);
    glNewList(camID, GL_COMPILE);
    cam->Draw();
    glEndList();

    // initialize camera frame transforms
    for (i=0; i < cameraCount; i++ )
    {
      double* c = cameras[i];
      wld2cam.push_back(FrameXform());
      glPushMatrix();
      glLoadIdentity();
      gluLookAt(c[0],c[1],c[2], c[3],c[4],c[5], c[6],c[7],c[8]);
      glGetDoublev( GL_MODELVIEW_MATRIX, wld2cam[i].matrix() );
      glPopMatrix();
      cam2wld.push_back(wld2cam[i].inverse());
    }
    cameraIndex = 0;
  }

  // set viewing transformation
  glLoadMatrixd(wld2cam[cameraIndex].matrix());

  // draw other cameras
  for (i=0; i < (int)wld2cam.size(); i++ )
  {
    if (i != cameraIndex)
    {
      glPushMatrix();
      glMultMatrixd(cam2wld[i].matrix());
      if (selectMode == 0)
      {
        drawFrame(5);
        float frontColor[] = {0.2, 0.2, 0.2, 1.0};
        glEnable(GL_LIGHTING);
        glMaterialfv(GL_FRONT, GL_AMBIENT, frontColor);
        glMaterialfv(GL_FRONT, GL_DIFFUSE, frontColor);
      }
      else
      {
        double r,g,b;       
        glDisable(GL_LIGHTING);
        munge(i+1, r,g,b);
        glColor3f(r, g, b);
      }
      glScaled(0.5,0.5,0.5);
      glTranslated(1.1,1.1,0.0);
      glCallList(camID);
      glPopMatrix();
    }
  }
}


//------------------------------------------------------------------------------
void drawFrame(float len)
{
  glDisable(GL_LIGHTING);
  glBegin(GL_LINES);
  glColor3d(1,0,0);
  glVertex3d(0,0,0);
  glVertex3d(len,0,0);
  glColor3d(0,1,0);
  glVertex3d(0,0,0);
  glVertex3d(0,len,0);
  glColor3d(0,0,1);
  glVertex3d(0,0,0);
  glVertex3d(0,0,len);
  glEnd();
}

//------------------------------------------------------------------------------
void drawCow()
{  
  if (frame == 0)
  {
    cow = new WaveFrontOBJ( "cow.obj" );
    cowID = glGenLists(1);
    glNewList(cowID, GL_COMPILE);
    cow->Draw();
    glEndList();
    glPushMatrix();
    glLoadIdentity();
    glTranslated(0,-cow->bbmin.y,-8);
    glRotated(-90, 0, 1, 0);
    glGetDoublev(GL_MODELVIEW_MATRIX, cow2wld.matrix());
    glPopMatrix();
  }
  glPushMatrix();
  glMultMatrixd(cow2wld.matrix());
  if (selectMode == 0)
  {
    drawFrame(5);
    float frontColor[] = {0.8, 0.2, 0.9, 1.0};
    glEnable(GL_LIGHTING);
    glMaterialfv(GL_FRONT, GL_AMBIENT, frontColor);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, frontColor);
  }
  else
  {
    double r,g,b;  
    glDisable(GL_LIGHTING);
    munge(32, r,g,b );
    glColor3d(r, g, b);    
  }
  glCallList(cowID);
  glPopMatrix();
}

//------------------------------------------------------------------------------
void drawBet()
{  
  if (frame == 0)
  {
    bet = new WaveFrontOBJ( "beethovan.obj" );
    betID = glGenLists(1);
    glNewList(betID, GL_COMPILE);
    bet->Draw();
    glEndList();
    glPushMatrix();
    glLoadIdentity();
    glTranslated(0,-bet->bbmin.y,8);
    glRotated(180, 0, 1, 0);
    glGetDoublev(GL_MODELVIEW_MATRIX, bet2wld.matrix());
    glPopMatrix();
  }
  glPushMatrix();
  glMultMatrixd(bet2wld.matrix());
  if (selectMode == 0)
  {
    drawFrame(8);
      float frontColor[] = {0.8, 0.3, 0.1, 1.0};
    glEnable(GL_LIGHTING);
    glMaterialfv(GL_FRONT, GL_AMBIENT, frontColor);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, frontColor);
  }
  else
  {
    double r,g,b;  
    glDisable(GL_LIGHTING);
    munge(33, r,g,b );
    glColor3d(r, g, b);    
  }
  glCallList(betID);
  glPopMatrix();
}
//------------------------------------------------------------------------------
void drawFloor()
{  
  if (frame == 0)
  {
    const int size = 8;
    unsigned char checker[size*size*3];
    for( int i=0; i < size*size; i++ )
    {
      if (((i/size) ^ i) & 1)
      {
        checker[3*i+0] = 200;
        checker[3*i+1] = 32;
        checker[3*i+2] = 32;
      }
      else
      {
        checker[3*i+0] = 200;
        checker[3*i+1] = 200;
        checker[3*i+2] = 32;
      }
    }
    glGenTextures( 1, &floorTexID );
    glBindTexture(GL_TEXTURE_2D, floorTexID);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glTexImage2D(GL_TEXTURE_2D, 0, 3, size, size, 0, GL_RGB, GL_UNSIGNED_BYTE, checker);
  }
  glDisable(GL_LIGHTING);
  if (selectMode == 0)
      glColor3d(0.35, .2, 0.1);
  else
  {
    double r,g,b;
    munge(34, r,g,b);
    glColor3d(r, g, b);
  }
  glBegin(GL_POLYGON);
  glVertex3f( 2000,-0.2, 2000);
  glVertex3f( 2000,-0.2,-2000);
  glVertex3f(-2000,-0.2,-2000);
  glVertex3f(-2000,-0.2, 2000);
  glEnd();
  if (selectMode == 0)
  {
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, floorTexID );
  }
  else
  {
    double r,g,b;
    munge(35, r,g,b);
    glColor3d(r, g, b);
  }
  glBegin(GL_POLYGON);
  glTexCoord2d(0,0);
  glVertex3d(-12,-0.1,-12);
  glTexCoord2d(1,0);
  glVertex3d( 12,-0.1,-12);
  glTexCoord2d(1,1);
  glVertex3d( 12,-0.1, 12);
  glTexCoord2d(0,1);
  glVertex3d(-12,-0.1, 12);
  glEnd();
  if (selectMode == 0)
  {
    glDisable(GL_TEXTURE_2D);
    drawFrame(5);
  }
}


//------------------------------------------------------------------------------
void display()
{
  if (selectMode == 0)
    glClearColor(0, 0.6, 0.8, 1);
  else
    glClearColor(0, 0, 0, 1);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  setCamera();
  drawFloor();
  drawCow();
  drawBet();
  glFlush();
  if (selectMode == 0)
      glutSwapBuffers();
  frame += 1;
}


//------------------------------------------------------------------------------
void reshape( int w, int h)
{
  width = w;
  height = h;
  glViewport(0, 0, width, height);
  glMatrixMode(GL_PROJECTION);            // Select The Projection Matrix
  glLoadIdentity();                       // Reset The Projection Matrix
  // Define perspective projection frustum
  double aspect = width/double(height);
  gluPerspective(45, aspect, 1, 1024);
  glMatrixMode(GL_MODELVIEW);             // Select The Modelview Matrix
  glLoadIdentity();                       // Reset The Projection Matrix
}

//------------------------------------------------------------------------------
void initialize()
{
  // Set up OpenGL state
  glShadeModel(GL_SMOOTH);         // Set Smooth Shading
  glEnable(GL_DEPTH_TEST);         // Enables Depth Testing
  glDepthFunc(GL_LEQUAL);          // The Type Of Depth Test To Do
  // Use perspective correct interpolation if available
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  // Initialize the matrix stacks
  reshape(width, height);
  // Define lighting for the scene
  float lightDirection[]   = {1.0, 1.0, 1.0, 0};
  float ambientIntensity[] = {0.1, 0.1, 0.1, 1.0};
  float lightIntensity[]   = {0.9, 0.9, 0.9, 1.0};
  glLightfv(GL_LIGHT0, GL_AMBIENT, ambientIntensity);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightIntensity);
  glLightfv(GL_LIGHT0, GL_POSITION, lightDirection);
  glEnable(GL_LIGHT0);
}

//------------------------------------------------------------------------------
void onMouseButton(int button, int state, int x, int y)
{
  y = height - y - 1;
  if (button == GLUT_LEFT_BUTTON)
  {
    if (state == GLUT_DOWN)
    {
      printf( "Left mouse click at (%d, %d)\n", x, y );
      selectMode = 1;
      display();
      glReadBuffer(GL_BACK);
      unsigned char pixel[3];
      glReadPixels(x, y, 1, 1, GL_RGB, GL_UNSIGNED_BYTE, pixel);
      printf( "pixel = %d\n", unmunge(pixel[0],pixel[1],pixel[2]));
      selectMode = 0;
      oldX = x;
      oldY = y;
    }
  }
  else if (button == GLUT_RIGHT_BUTTON)
  {
    printf( "Right mouse click at (%d, %d)\n",x,y );
  }
  glutPostRedisplay();
}


//------------------------------------------------------------------------------
void onMouseDrag( int x, int y)
{
  y = height - y - 1;
  printf( "in drag (%d, %d)\n", x - oldX,  y - oldY );
  glutPostRedisplay();
}

//------------------------------------------------------------------------------
void onKeyPress( unsigned char key, int x, int y)
{
  if ((key == ' ') || (key == 'c'))
  {    
    printf( "Toggle camera %d\n", cameraIndex );
    cameraIndex += 1;
  }      
  else if ((key >= '0') && (key <= '9'))
    cameraIndex = key - '0';
  if (cameraIndex >= (int)wld2cam.size() )
      cameraIndex = 0;
  glutPostRedisplay();
}

//------------------------------------------------------------------------------
void main( int argc, char* argv[] )
{
  width = 800;
  height = 600;
  frame = 0;
  glutInit( &argc, argv ); 
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
  glutInitWindowSize(width, height);
  glutInitWindowPosition(100, 100);
  glutCreateWindow("Simple Scene");
  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(onKeyPress);
  glutMouseFunc(onMouseButton);
  glutMotionFunc(onMouseDrag);
  int rv,gv,bv;
  glGetIntegerv(GL_RED_BITS,&rv);
  glGetIntegerv(GL_GREEN_BITS,&gv);
  glGetIntegerv(GL_BLUE_BITS,&bv);
  printf( "Pixel depth = %d : %d : %d\n", rv, gv, bv );
  initialize();
  glutMainLoop();
}
